/*
MIT License

Copyright (c) 2018 Janos Buttgereit

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#pragma once

#include <JuceHeader.h>
#include <mutex>

class WindowOpenGLContext : private juce::OpenGLRenderer
{
public:
    template<typename OpenGLRendererType>
    struct OpenGLRendererOwner
    {
        static_assert(std::is_base_of<juce::OpenGLRenderer, OpenGLRendererType>::value);
        static_assert(std::is_base_of<juce::Component, OpenGLRendererType>::value);

        OpenGLRendererOwner(std::shared_ptr<OpenGLRendererType> renderer, WindowOpenGLContext& context) : context(context)
        {
            this->renderer = renderer;
            jassert(renderer != nullptr);

            context.addRenderingTarget(renderer);
        }

        ~OpenGLRendererOwner()
        {
            context.removeRenderingTargetAsync(renderer);
        }

        OpenGLRendererType* get()
        {
            return renderer.get();
        }

    private:
        WindowOpenGLContext& context;
        std::shared_ptr<OpenGLRendererType> renderer;
    };

    WindowOpenGLContext();

    ~WindowOpenGLContext() override;

    void setTopLevelParentComponent(juce::Component& topLevelComponent);

    void detachTopLevelParentComponent();

    void executeOnGLThread(std::function<void(juce::OpenGLContext & )>&& lambdaToExecute);

    void executeOnGLThreadMultipleTimes(std::function<void(juce::OpenGLContext & )>&& lambdaToExecute, const int repetitions);

    juce::Rectangle<int> getComponentClippingBoundsRelativeToGLRenderingTarget(juce::Component* targetComponent);

    juce::OpenGLContext openGLContext;

private:
    juce::Component* topLevelComponent = nullptr;

    std::vector<std::shared_ptr<juce::OpenGLRenderer>> renderingTargets;
    std::mutex renderingTargetsLock;

    std::vector<std::function<void(juce::OpenGLContext & )>> executeInRenderCallback;
    std::mutex executeInRenderCallbackLock;

    std::mutex releasePoolLock;
    std::vector<std::shared_ptr<juce::OpenGLRenderer>> rendererTargetsReleasePool;

    bool addRenderingTarget(std::shared_ptr<juce::OpenGLRenderer> newTarget);

    void removeRenderingTargetAsync(std::shared_ptr<juce::OpenGLRenderer> targetToRemove);

    void cleanupRendererReleasePool();

    // OpenGL related member functions
    void newOpenGLContextCreated() override;

    void renderOpenGL() override;

    void openGLContextClosing() override;

    JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (WindowOpenGLContext)
};
